/*********************************************************************
  UCMD.C:	
	Fuehrt den angegebenen UNIX-Befehl fuer alle Benutzer aus.
	UCMD bietet dabei fuer den Befehl folgende Variablen:
		#user	Benutzername aus /etc/passwd
		#group	dazugehoeriger Gruppenname aus /etc/passwd
		#home	dazugehoeriges Homeverzeichnis aus /etc/passwd

  Projekt:	
	ucmd

  Funktionen:
	main
 ---------------------------------------------
  Erstellt:   08/92, M. Trapp
 ---------------------------------------------
  Datum     Aenderung						 von
**********************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pwd.h>
#include <grp.h>

/*--- Konstanten ------*/
#define N_NULL	(NUTZER *) 0
#define SYSV
#define NAMELEN	20
#define CMDLEN  200		/* Laenge des Befehls mit Variablen und aufgeloest */

/*--- Typdefinitionen ---*/
typedef struct nutzer
{
	char user[NAMELEN];			/* Loginname des Benutzers, kleingeschrieben */
	char group[NAMELEN];		/* Gruppennamen oder GID als Zahl */
	char home[1024];			/* Heimatverzeichnis des Benutzers */
	struct nutzer *next;
} NUTZER;

/*--- Funktionsprototypen ------*/
void Hilfe();
int HoleNamen();
char *ErsetzeVars();
void ZeigeNamen();
void SortiereNamen();
void ZieheVor();
#ifdef SYSV
void bzero();
#endif
char *strlwr();

main (argc, argv)
int argc;
char **argv;
{
	NUTZER *ohne;
	NUTZER *nutzer;
	NUTZER *neu, *o, *n, *v;
	char name[NAMELEN];
	char rohbef[CMDLEN];
	char *befehl;
	int verbose;
	int ShowList, DryRun;
	FILE *k;

	/* Initialisierung: */
	ohne = nutzer = neu = N_NULL;
	verbose = 1;
	ShowList = DryRun = 0;

	while ( (argc > 1) && (argv[1][0] == '-') )
	{
		switch ( toupper (argv[1][1]) )
		{
		  case 'H':
		  default:
			Hilfe();
			return (1);

		  case 'Q':		/* Zeige nicht, welcher User bearbeitet wird */
			verbose = 0;
			break;

		  case 'N':  /* ohne die User, die in der folgenden Datei stehen */
			if ( argc < 3 )
			{
			  printf("ucmd: Die Dateiangabe fehlt fuer Option -n!\n");
			  return (2);
			}
			if ( (k = fopen (argv[2], "rt")) == NULL )
			{
				printf ("ucmd: Die Datei %s ist nicht vorhanden!\n", argv[2]);
				return (3);
			}

			/* Datei mit Usernamen einlesen: */
			while ( fscanf (k, "%s", name) == 1 )
			{
				strlwr (name);
				if ( (neu = (NUTZER *) malloc ( sizeof (NUTZER))) == N_NULL )
				{
					printf ("ucmd: zuwenig Speicher frei!\n");
					return (4);
				}
				strcpy (neu->user, name);
				neu->group[0] = '\0';
				neu->home[0] = '\0';
				/* in die Liste stellen */
				if ( ohne == N_NULL )
				{
					neu->next = N_NULL;
					ohne = neu;
				}
				else
				{
					neu->next = ohne;
					ohne = neu;
				}
			}  /* solange, bis die Datei eingelesen wurde */

			ZieheVor (&argc, argv);
			break;

		case 'S':  /* zeige nur die Liste aller angemeldeten User.
					  Mit Option -N die verkuerzte Liste. */
			ShowList = 1;
			break;

		case 'D':  /* zeige, welcher Befehl ausgefuehrt wuerde (Dry run) */
			DryRun = 1;
			break;

		}  /* Fallunterscheidung Optionen */
		ZieheVor (&argc, argv);
	}  /* 2. Argument hatte Strich */

	printf ("\nUCMD V1.0:\n");

	/* Benutzernamen und -gruppen lesen: */
	if ( HoleNamen (&nutzer) != 0 )
		return (5);	 /* Zuwenig Speicherplatz */

#ifdef DEBUG
	printf ("Nach HoleNamen() und Einlesen der Ausschlussdatei:\n");
	ZeigeNamen (ohne);
	ZeigeNamen (nutzer);
#endif

	/* Namensliste um die nichtgewuenschten Benutzer kuerzen: */
	for ( o = ohne; o != N_NULL; o = o->next )
	{
		for ( n = nutzer; n != N_NULL; n = n->next )
		{
			if ( strcmp (o->user, n->user) == 0 )
			{
				/* Eintrag aus Liste loeschen: */
				if ( n == nutzer )
					nutzer = nutzer->next;
				else
				{
					for ( v = nutzer; v->next != n; v = v->next);
					v->next = n->next;
				}
				free (n);
				break;
			}  /* solange, bis alle User verglichen wurden */
		}  /* dieser User ist nicht erwuenscht! */
	}  /* solange, bis alle nicht gewuenschten durch sind */

#ifdef DEBUG
	printf ("Nach dem Vergleich mit OHNE:\n");
	ZeigeNamen (ohne);
	ZeigeNamen (nutzer);
#endif

	if ( ShowList )
	{
		ZeigeNamen (nutzer);
		return (0);
	}  /* nur Liste der Namen gewuenscht */

	if ( argc != 2 )
	{
		printf ("ucmd: der Befehl fehlt!\n");
		return (4);
	}

	/* verkuerzte Liste der Benutzer durchgehen und jedesmal den Befehl
	   ausfuehren: */
	for ( n = nutzer; n != N_NULL; n = n->next )
	{
		if ( verbose )
			printf ("\nWende den Befehl auf '%s' an ...\n", n->user);

		/* Evt. Variablen in der Befehlszeile ersetzen: */
		strcpy (rohbef, argv[1]);
		befehl = ErsetzeVars (rohbef, n);

#ifndef DEBUG
		if ( DryRun )
			printf ("\t%s ... dryrun.\n", befehl);
		else
		{
			/* Der Befehl system() ruft die Bourne-Shell auf.
			   Die C-Shell kann man nicht durch die exec...()-Funktionen
			   aufrufen, da diese das C-Programm terminieren.
			*/
			system (befehl);
#else
			printf ("Der Befehl lautet %s\n", befehl);
#endif
		}  /* Befehl wirklich ausfuehren, kein DryRun */

	};
	printf ("\n");

	return (0);
}  /* main() */


#ifdef SYSV
void bzero(b, laenge)
char *b;
int laenge;
{
int i;
 for(i=0; i <= laenge; i++)
     *b++ = (char) '\0';
}
#endif


void Hilfe()
{
	printf ("\n\nucmd V1.0:\tUser Command fuehrt einen Befehl fuer jeden Benutzer aus\n\n");
	printf ("Aufruf:\t\tucmd [-H] [-Q] [-N datei] [-S] [-D] befehl\n\n");
	printf ("Optionen:\n");
	printf ("  -H\t\tHelp\n");
	printf ("  -Q\t\tQuite\n");
	printf ("  -N datei\tnot the user in this file\n");
	printf ("  -S\t\tShow userlist\n");
	printf ("  -D\t\tDry Run\n\n");
	printf ("Hinweis:\n");
	printf ("  Fuer den Befehl stehen folgende Variablen zur Verfuegung:\n");
	printf ("\t#user\t\tBenutzername aus der Datei /etc/passwd\n");
	printf ("\t#group\t\tzugehoerige Gruppe aus /etc/passwd\n");
	printf ("\t#home\t\tzugehoeriges Heimatverzeichnis aus /etc/passwd\n\n\n");
}  /* Hilfe() */


int HoleNamen (nutzer)
/*=========================================================================
  Beschreibung:  
	Liest jede Zeile der Datei /etc/passwd und holt die Benutzer heraus,
	deren UID zwischen 100 (einschliesslich) und 499 liegt.
	Der Benutzername wird nach nutzer->user, das Heimatverzeichnis nach
	nutzer->home kopiert.
	Dann wird versucht zur zugeh. GID den Namen der Gruppe aus /etc/group
	zu lesen. Falls dies nicht gelingt, wird in nutzer->group die GID als
	Zahl eingetragen.
	Die Benutzer des Systems werden in einer einfach verketteten, unsortierten
	Kette verwaltet.

  Rueckgabewerte:  
	nutzer:	verkettete Liste aller gefundenen Benutzer (s.o.)
	0, falls alles in Ordnung ist.
	6, falls zuwenig Speicher da ist.

  Anmerkungen:  
	keine

  Globale Variablen:  
	keine
  -------------------------------------------------------------------------
  Datum     Aenderung                                          von
===========================================================================*/
NUTZER **nutzer;
{
	struct passwd *line;
	struct group *gline;
	NUTZER *neu;

	*nutzer = N_NULL;
	while ( (line = getpwent()) != NULL )	
	{
		if ( (line->pw_uid >= 100) && (line->pw_uid < 500) )
		{
			if ( (neu = (NUTZER *) malloc (sizeof (NUTZER))) == N_NULL )
			{
				printf ("ucmd: zuwenig Speicher!\n");
				return (6);
			}
			strcpy ( neu->user, line->pw_name);
			strlwr (neu->user);

			/* Gruppennamen aus /etc/group holen */
			if ( (gline = getgrgid (line->pw_gid)) == NULL )
				sprintf (neu->group, "%d", line->pw_gid);
			else
				strcpy (neu->group, gline->gr_name);

			strcpy ( neu->home, line->pw_dir);
			if ( *nutzer == N_NULL )
			{
				neu->next = N_NULL;
				*nutzer = neu;
			}
			else
			{
				neu->next = *nutzer;
				*nutzer = neu;
			}
		}  /* UID ist groesser als 100 */
	}  /* solange, bis alle Eintraege durch sind */
	endpwent();
	return (0);
}  /* HoleNamen() */


char *ErsetzeVars (zeile, nutzer)
/*=========================================================================
  Beschreibung:  
	Ersetzt alle Token #user durch nutzer->user, #group durch nutzer->
	group und #home durch nutzer->home.
	Der Puffer, auf den zeile zeigt wird veraendert. Der Rueckgabewert
	zeigt auf einen statischen Puffer, der bei jedem Aufruf wiederver-
	wendet wird.

  Rueckgabewerte:  
	Umgewandelte Zeile.

  Anmerkungen:
	siehe oben

  Globale Variablen:  
	keine
  -------------------------------------------------------------------------
  Datum     Aenderung                                          von
===========================================================================*/
char *zeile;		/* Befehlszeile mit Variablen */
NUTZER *nutzer;		/* Benutzer, fuer die die Variablen ersetzt werden sollen */
{
	static char ziel[CMDLEN];
	char *gatter, *vorher, *rest;

	rest = vorher = zeile;
	bzero (ziel, CMDLEN);

	while ( gatter = strchr (rest, '#') )
	{
		if ( strncmp (gatter+1, "user", 4) == 0 )
		{
			strncat (ziel, vorher, gatter - vorher);
			strcat (ziel, nutzer->user);
			vorher = gatter + 5;
		}
		rest = gatter + 1;
	}  /* ersetze #user */
	strcat (ziel, vorher);

	strcpy (zeile, ziel);
	rest = vorher = zeile;
	bzero (ziel, CMDLEN);

	while ( gatter = strchr (rest, '#') )
	{
		if ( strncmp (gatter+1, "group", 5) == 0 )
		{
			strncat (ziel, vorher, gatter - vorher);
			strcat (ziel, nutzer->group);
			vorher = gatter + 6;
		}
		rest = gatter + 1;
	}  /* ersetze #group */
	strcat (ziel, vorher);

	strcpy (zeile, ziel);
	rest = vorher = zeile;
	bzero (ziel, CMDLEN);

	while ( gatter = strchr (rest, '#') )
	{
		if ( strncmp (gatter+1, "home", 4) == 0 )
		{
			strncat (ziel, vorher, gatter - vorher);
			strcat (ziel, nutzer->home);
			vorher = gatter + 5;
		}
		rest = gatter + 1;
	}  /* ersetze #home */
	strcat (ziel, vorher);

	return (ziel);
}  /* ErsetzeVars() */



void ZeigeNamen (liste)
/*=========================================================================
  Beschreibung:  
	Zeigt die alphabetisch nach Loginnamen sortierte Liste der Benutzer
	mit Gruppenzugehoerigkeit und Heimatverzeichnis an.

  Rueckgabewerte:  
	Keine.

  Anmerkungen:  
	Keine.

  Globale Variablen:  
	keine.
  -------------------------------------------------------------------------
  Datum     Aenderung                                          von
===========================================================================*/
NUTZER *liste;
{
	NUTZER *akt;

	/* SortiereNamen (&liste);  +++ noch nicht fertig +++ */
	printf ("Name\t\tGruppe\t\tHeimatverzeichnis\n");
	printf ("----------------------------------------------------------------------\n");
	for (akt=liste; akt != N_NULL; akt = akt->next )
		printf ("%s\t\t%s\t\t%s\n", akt->user, akt->group, akt->home);
	printf ("\n");
}  /* ZeigeNamen() */


void SortiereNamen (liste)
/*=========================================================================
  Beschreibung:  
	Sortiert die Liste alphabetisch nach Loginnamen. 
	Der Selectionsort-Algorithmus ist dem Buch "R. Sedgewick: Algorithmen"
	entnommen und wurde fuer einfach verkettete Listen angepasst.

  Rueckgabewerte:  
	liste	sortierte Liste.

  Anmerkungen:  
	Funktion erzeugt einen Bus-Error --> austesten!

  Globale Variablen:  
	keine.
  -------------------------------------------------------------------------
  Datum     Aenderung                                          von
===========================================================================*/
NUTZER **liste;
{
	NUTZER *i, *j, *vori, *vorj, *min, *vormin, *t;

	if ( *liste == N_NULL )
		return;

	for ( vori = N_NULL, i = *liste;
		  i->next != N_NULL;
		  vori = i, i = i->next 
		)
	{
		min = i;
		vormin = vori;
		for ( vorj = i, j = i->next;
			  j != N_NULL;
			  vorj = j, j = j->next
			)
			if ( strcmp (j->user, min->user) < 0 )
			{
				vormin = vorj;
				min = j;
			}  /* neues Minimum gefunden */

		/* Minimum mit i. Element vertauschen: */
		if ( vori == N_NULL )
			*liste = min;
		else
			vori->next = min;
		t = min->next;
		min->next = i->next;
		if ( vormin != N_NULL )
			vormin->next = i;
		i->next = t;
	}  /* solange, bis Liste durchgesehen ist */
}  /* SortiereNamen() */


void ZieheVor (argc, argv)
/*=========================================================================
  Beschreibung: 
	Kopiert den Zeiger des dritten Eintrags auf den zweiten,
	den des vierten auf den dritten und so fort, bis das
	ganze Feld um eine Stelle nach vorne gezogen wurde.
	In den letzten, nun freie Eintrag, wird der Nullzeiger
	geschrieben. Die Anzahl der Eintraege (argc) wird um
	eins erniedrigt.
  Rueckgabewerte    : keine!
  Speicheranf.	   : keine!
  Ausnahmen	   : keine!
  Fehlermeldung    : keine!
  Globale Variablen: keine!
  Funktionsaufrufe : keine!
 -------------------------------------------------------
  Datum     Aenderung						 von
=========================================================================*/
int *argc;
char **argv;
{
  int i;

  (*argc)--;
  for (i=1; i < *argc; i++)
    argv[i] = argv[i+1];
  argv[*argc] = NULL;
  return;
}  /* ZieheVor() */



char *strlwr (Text)
/*=========================================================================
  Beschreibung: 
	Konvertiert den angeg. Text in Kleinbuchstaben.

  Rueckgabewerte: 
	Zeiger auf den Puffer, der den umgewandelten Text enthaelt.

  Anmerkungen:
	keine.

  Globale Variablen: 
 	keine.
 -------------------------------------------------------
  Datum     Aenderung						 von
=========================================================================*/
char *Text;
{
	char *Temp;

	Temp = Text;
	while ( *Temp )
	{
		*Temp = tolower (*Temp);
		Temp++;
	};
	return (Text);
}  /* strlwr() */

/*-------- EOF --------- EOF ----------- EOF --------------------*/
